# To obtain different panels of Figure 1 in the paper, search for "choose one of:"
#   and make appropriate changes.
import math
from numpy import linalg
from numpy import dot
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
def true(x1,x2):
  """the true kernel"""
  return math.cos(x1[0]-x2[0])+math.cos(x1[1]-x2[1])
def laplace(x1,x2):
  """Laplacian kernel"""
  return math.exp(-linalg.norm(x1-x2))
def linear(x1,x2):
  """linear kernel"""
  return dot(x1,x2)
kernel = true        # the kernel to use (choose one of: true, laplace, linear)
n = 1000             # the size of the training set
x_test = np.ones(2)  # the test object
a = 1                # ridge constant
np.random.seed(0)    # for reproducibility and comparability
x = 2*np.random.random([n,2])-1  # the training attributes
w = np.random.randn(4)           # the weights
y = np.zeros(n)         # initializing the training labels
for i in range(n):      # the training labels
  y[i] = w[0]*math.cos(x[i,0]) + w[1]*math.cos(x[i,1]) + w[2]*math.sin(x[i,0]) + w[3]*math.sin(x[i,1]) \
         + np.random.randn()   # Gaussian noise
K = np.zeros((n,n))     # initializing the kernel matrix
H = np.zeros((n,n))     # initializing the hat matrix
H_diag = np.zeros(n)    # initializing the diagonal of the hat matrix
k = np.zeros(n)         # initializing the kernel vector
for i in range(n):
  for j in range(n):
    K[i,j] = kernel(x[i],x[j])       # the kernel matrix (step 1 of preprocessing in the paper)
K_inv = linalg.inv(K+a*np.eye(n))    # step 2 of preprocessing in the paper
H = dot(K_inv,K)                     # the hat matrix (step 3 of preprocessing in the paper)
for i in range(n):
  H_diag[i] = H[i,i]                 # the hat matrix diagonal (step 3 of preprocessing in the paper)
y_hat = dot(dot(K_inv,K),y)          # y hat  (step 4 of preprocessing in the paper)
for i in range(n):
  k[i] = kernel(x[i],x_test)               # the kernel vector (prediction step 1 in the paper)
kappa = kernel(x_test,x_test)              # the kernel scalar (prediction step 1 in the paper)
epsilon = dot(K_inv,k)                     # prediction step 2 in the paper
semi_d = dot(k,epsilon)                    # prediction step 3 in the paper
d = 1/(kappa+a-semi_d)                     # prediction step 4 in the paper
one_minus_h = 1-H_diag+a*d*(epsilon**2)    # prediction step 5 in the paper
hat_test = dot(y,epsilon)                  # prediction step 6 in the paper
sqrt_one_minus_h = np.zeros(n)   # initializing
for i in range(n):
  sqrt_one_minus_h[i] = math.sqrt(one_minus_h[i])
A = math.sqrt(a*d)*hat_test + (y-y_hat+a*d*hat_test*epsilon)/sqrt_one_minus_h
B = math.sqrt(a*d)+(a*d*epsilon)/sqrt_one_minus_h
C = A/B
C_srt = np.sort(C)
left_side = -4   # left boundary of the plot
right_side = 2   # right boundary of the plot
plt.plot([left_side,C_srt[0]],[0,0],'b')
plt.plot([left_side,C_srt[0]],[1/(n+1),1/(n+1)],'b')
for i in range(n-1):
  plt.plot([C_srt[i],C_srt[i]],[i/(n+1),(i+2)/(n+1)],'b')
  plt.plot([C_srt[i],C_srt[i+1]],[(i+1)/(n+1),(i+1)/(n+1)],'b')
  plt.plot([C_srt[i],C_srt[i+1]],[(i+2)/(n+1),(i+2)/(n+1)],'b')
plt.plot([C_srt[n-1],C_srt[n-1]],[(n-1)/(n+1),1],'b')
plt.plot([C_srt[n-1],right_side],[n/(n+1),n/(n+1)],'b')
plt.plot([C_srt[n-1],right_side],[1,1],'b')
Bayes_spread = math.sqrt(1 + kappa - dot(k,epsilon))
range_cdf = np.linspace(left_side,right_side)
plt.plot(range_cdf,norm.cdf(range_cdf,loc=hat_test,scale=Bayes_spread),'r')
plt.show()
